/*****
*
* Hello SQLite World! example
* $Id: hello.js 107 2008-08-20 08:30:20Z klaus $
*
*****/

window.onload = function() {
    // connect to database within same directory as .svg file
    window.CONN = openDatabase(document.URL.replace("hello.svg","hello.db"));

    // get reference to text-element and add eventListener
    var textNode = document.getElementById('hello.msg');
    textNode.onclick = function() {
        var msg = prompt("Enter message",textNode.firstChild.nodeValue);    // ask for new text
        if (msg && !msg.match(/^ +$/)) {
            executeSql("UPDATE hello SET msg=?1",[msg]);    // update database
            queryMessage();   // display message
        }
    };

    // query current message from database and set text in SVG
    queryMessage = function() {
        executeSql("CREATE TABLE IF NOT EXISTS hello (msg TEXT)");
        var res = executeSql("SELECT msg FROM hello");
        if (res.length == 1) {
            textNode.firstChild.nodeValue = res[0]['msg'];
        }
        else {
            // table hello is empty so populate it
            executeSql("INSERT INTO hello (msg) VALUES ('Hello SQLite World!')");
            textNode.firstChild.nodeValue = 'Hello SQLite World!';
        }
    };

    // get current message from database onload
    queryMessage();
}

function openDatabase(dbURI) {
    try {
        // grant enhanced abilities to script
        netscape.security.PrivilegeManager.enablePrivilege("UniversalXPConnect");

        // open local database and return db-handle
        var ioService = Components.classes["@mozilla.org/network/io-service;1"].
            getService(Components.interfaces.nsIIOService);
        var dbFile = ioService.newURI(dbURI,null,null).
            QueryInterface(Components.interfaces.nsIFileURL).file;
        var stService = Components.classes["@mozilla.org/storage/service;1"].
            getService(Components.interfaces.mozIStorageService);
        return stService.openDatabase(dbFile);
    }
    catch(err) {
        alert('Error: Could not connect to database '+dbURI+'.\nStatus: '+err);
    }
}

function executeSql(sqlStatement,sqlParams) {
    // grant enhanced abilities to script
    netscape.security.PrivilegeManager.enablePrivilege("UniversalXPConnect");

    // create statement, bind parameters, execute it and return result-set as dictionary
    var res = [];
    var stmt = window.CONN.createStatement(sqlStatement);
    if (sqlParams) {
        for (var i=0; i<sqlParams.length; i++) {
            stmt.bindUTF8StringParameter(i,sqlParams[i]);
        }
    }
    while (stmt.executeStep()) {
        var row = {};
        for (var i=0; i<stmt.numEntries;i++) {
            var colName = stmt.getColumnName(i).toLowerCase();
            row[colName] = stmt.getUTF8String(i);
        }
        res[res.length] = row;
    }
    return res;
}
